# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: grpc_helper/api/events.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from grpc_helper.api import common_pb2 as grpc__helper_dot_api_dot_common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
    name="grpc_helper/api/events.proto",
    package="",
    syntax="proto3",
    serialized_options=b"\n\017grpc_helper.apiP\001",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1cgrpc_helper/api/events.proto\x1a\x1cgrpc_helper/api/common.proto",\n\rEventProperty\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t"9\n\x05\x45vent\x12\x0c\n\x04name\x18\x01 \x01(\t\x12"\n\nproperties\x18\x02 \x03(\x0b\x32\x0e.EventProperty"K\n\x0b\x45ventStatus\x12\x12\n\x01r\x18\x01 \x01(\x0b\x32\x07.Result\x12\x15\n\x05\x65vent\x18\x02 \x01(\x0b\x32\x06.Event\x12\x11\n\tclient_id\x18\x03 \x01(\r"/\n\x0b\x45ventFilter\x12\r\n\x05names\x18\x01 \x03(\t\x12\x11\n\tclient_id\x18\x02 \x01(\r"#\n\x0e\x45ventInterrupt\x12\x11\n\tclient_id\x18\x01 \x01(\r":\n\x10\x45ventQueueStatus\x12\x12\n\x01r\x18\x01 \x01(\x0b\x32\x07.Result\x12\x12\n\nclient_ids\x18\x02 \x03(\r*X\n\x0f\x45ventApiVersion\x12\x15\n\x11\x45VENT_API_UNKNOWN\x10\x00\x12\x17\n\x13\x45VENT_API_SUPPORTED\x10\x01\x12\x15\n\x11\x45VENT_API_CURRENT\x10\x02\x32\xa8\x01\n\x0c\x45ventService\x12&\n\x06listen\x12\x0c.EventFilter\x1a\x0c.EventStatus0\x01\x12+\n\tinterrupt\x12\x0f.EventInterrupt\x1a\r.ResultStatus\x12\x1d\n\x04send\x12\x06.Event\x1a\r.ResultStatus\x12$\n\x07inspect\x12\x06.Empty\x1a\x11.EventQueueStatusB\x13\n\x0fgrpc_helper.apiP\x01\x62\x06proto3',
    dependencies=[
        grpc__helper_dot_api_dot_common__pb2.DESCRIPTOR,
    ],
)

_EVENTAPIVERSION = _descriptor.EnumDescriptor(
    name="EventApiVersion",
    full_name="EventApiVersion",
    filename=None,
    file=DESCRIPTOR,
    create_key=_descriptor._internal_create_key,
    values=[
        _descriptor.EnumValueDescriptor(
            name="EVENT_API_UNKNOWN", index=0, number=0, serialized_options=None, type=None, create_key=_descriptor._internal_create_key
        ),
        _descriptor.EnumValueDescriptor(
            name="EVENT_API_SUPPORTED", index=1, number=1, serialized_options=None, type=None, create_key=_descriptor._internal_create_key
        ),
        _descriptor.EnumValueDescriptor(
            name="EVENT_API_CURRENT", index=2, number=2, serialized_options=None, type=None, create_key=_descriptor._internal_create_key
        ),
    ],
    containing_type=None,
    serialized_options=None,
    serialized_start=390,
    serialized_end=478,
)
_sym_db.RegisterEnumDescriptor(_EVENTAPIVERSION)

EventApiVersion = enum_type_wrapper.EnumTypeWrapper(_EVENTAPIVERSION)
EVENT_API_UNKNOWN = 0
EVENT_API_SUPPORTED = 1
EVENT_API_CURRENT = 2


_EVENTPROPERTY = _descriptor.Descriptor(
    name="EventProperty",
    full_name="EventProperty",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="EventProperty.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="EventProperty.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=62,
    serialized_end=106,
)


_EVENT = _descriptor.Descriptor(
    name="Event",
    full_name="Event",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="Event.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="properties",
            full_name="Event.properties",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=108,
    serialized_end=165,
)


_EVENTSTATUS = _descriptor.Descriptor(
    name="EventStatus",
    full_name="EventStatus",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="r",
            full_name="EventStatus.r",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="event",
            full_name="EventStatus.event",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="client_id",
            full_name="EventStatus.client_id",
            index=2,
            number=3,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=167,
    serialized_end=242,
)


_EVENTFILTER = _descriptor.Descriptor(
    name="EventFilter",
    full_name="EventFilter",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="names",
            full_name="EventFilter.names",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="client_id",
            full_name="EventFilter.client_id",
            index=1,
            number=2,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=244,
    serialized_end=291,
)


_EVENTINTERRUPT = _descriptor.Descriptor(
    name="EventInterrupt",
    full_name="EventInterrupt",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="client_id",
            full_name="EventInterrupt.client_id",
            index=0,
            number=1,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=293,
    serialized_end=328,
)


_EVENTQUEUESTATUS = _descriptor.Descriptor(
    name="EventQueueStatus",
    full_name="EventQueueStatus",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="r",
            full_name="EventQueueStatus.r",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="client_ids",
            full_name="EventQueueStatus.client_ids",
            index=1,
            number=2,
            type=13,
            cpp_type=3,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=330,
    serialized_end=388,
)

_EVENT.fields_by_name["properties"].message_type = _EVENTPROPERTY
_EVENTSTATUS.fields_by_name["r"].message_type = grpc__helper_dot_api_dot_common__pb2._RESULT
_EVENTSTATUS.fields_by_name["event"].message_type = _EVENT
_EVENTQUEUESTATUS.fields_by_name["r"].message_type = grpc__helper_dot_api_dot_common__pb2._RESULT
DESCRIPTOR.message_types_by_name["EventProperty"] = _EVENTPROPERTY
DESCRIPTOR.message_types_by_name["Event"] = _EVENT
DESCRIPTOR.message_types_by_name["EventStatus"] = _EVENTSTATUS
DESCRIPTOR.message_types_by_name["EventFilter"] = _EVENTFILTER
DESCRIPTOR.message_types_by_name["EventInterrupt"] = _EVENTINTERRUPT
DESCRIPTOR.message_types_by_name["EventQueueStatus"] = _EVENTQUEUESTATUS
DESCRIPTOR.enum_types_by_name["EventApiVersion"] = _EVENTAPIVERSION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EventProperty = _reflection.GeneratedProtocolMessageType(
    "EventProperty",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENTPROPERTY,
        "__module__": "grpc_helper.api.events_pb2"
        # @@protoc_insertion_point(class_scope:EventProperty)
    },
)
_sym_db.RegisterMessage(EventProperty)

Event = _reflection.GeneratedProtocolMessageType(
    "Event",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENT,
        "__module__": "grpc_helper.api.events_pb2"
        # @@protoc_insertion_point(class_scope:Event)
    },
)
_sym_db.RegisterMessage(Event)

EventStatus = _reflection.GeneratedProtocolMessageType(
    "EventStatus",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENTSTATUS,
        "__module__": "grpc_helper.api.events_pb2"
        # @@protoc_insertion_point(class_scope:EventStatus)
    },
)
_sym_db.RegisterMessage(EventStatus)

EventFilter = _reflection.GeneratedProtocolMessageType(
    "EventFilter",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENTFILTER,
        "__module__": "grpc_helper.api.events_pb2"
        # @@protoc_insertion_point(class_scope:EventFilter)
    },
)
_sym_db.RegisterMessage(EventFilter)

EventInterrupt = _reflection.GeneratedProtocolMessageType(
    "EventInterrupt",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENTINTERRUPT,
        "__module__": "grpc_helper.api.events_pb2"
        # @@protoc_insertion_point(class_scope:EventInterrupt)
    },
)
_sym_db.RegisterMessage(EventInterrupt)

EventQueueStatus = _reflection.GeneratedProtocolMessageType(
    "EventQueueStatus",
    (_message.Message,),
    {
        "DESCRIPTOR": _EVENTQUEUESTATUS,
        "__module__": "grpc_helper.api.events_pb2"
        # @@protoc_insertion_point(class_scope:EventQueueStatus)
    },
)
_sym_db.RegisterMessage(EventQueueStatus)


DESCRIPTOR._options = None

_EVENTSERVICE = _descriptor.ServiceDescriptor(
    name="EventService",
    full_name="EventService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=481,
    serialized_end=649,
    methods=[
        _descriptor.MethodDescriptor(
            name="listen",
            full_name="EventService.listen",
            index=0,
            containing_service=None,
            input_type=_EVENTFILTER,
            output_type=_EVENTSTATUS,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="interrupt",
            full_name="EventService.interrupt",
            index=1,
            containing_service=None,
            input_type=_EVENTINTERRUPT,
            output_type=grpc__helper_dot_api_dot_common__pb2._RESULTSTATUS,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="send",
            full_name="EventService.send",
            index=2,
            containing_service=None,
            input_type=_EVENT,
            output_type=grpc__helper_dot_api_dot_common__pb2._RESULTSTATUS,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="inspect",
            full_name="EventService.inspect",
            index=3,
            containing_service=None,
            input_type=grpc__helper_dot_api_dot_common__pb2._EMPTY,
            output_type=_EVENTQUEUESTATUS,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_EVENTSERVICE)

DESCRIPTOR.services_by_name["EventService"] = _EVENTSERVICE

# @@protoc_insertion_point(module_scope)



import copy
import os
import socket
import time
from logging import Logger, getLogger
from typing import TypeVar, Union

from grpc import RpcError, StatusCode, insecure_channel

from grpc_helper.api import Result, ResultCode
from grpc_helper.errors import RpcException
from grpc_helper.meta import RpcMetadata
from grpc_helper.static_config import RPC_RETRY_DELAY
from grpc_helper.utils import get_current_ip, is_streaming, trace_rpc


class RetryMethod:
    def __init__(self, name: str, stub, channel: str, timeout: float, metadata: RpcMetadata, logger: Logger, exception: bool, custom_exception: TypeVar):
        self.m_name = name
        self.s_name = f"{stub.__class__.__name__}({channel})"
        self.stub = stub
        self.timeout = timeout
        self.metadata = metadata
        self.logger = logger
        self.exception = exception
        self.custom_exception = custom_exception if custom_exception is not None else RpcException

    def prelude(self, request):
        trace = trace_rpc(True, request, context=self.metadata, method=f"{self.s_name}.{self.m_name}")
        self.logger.debug(trace)
        return (trace, time.time())

    def handle_exception(self, trace: str, first_try: float, e: RpcError, retry_delay: float):
        if e.code() == StatusCode.UNAVAILABLE and self.timeout is not None and (time.time() - first_try) < self.timeout:
            # Server is not available, and timeout didn't expired yet: sleep and retry
            self.logger.debug(f"<RPC> << {self.s_name}.{self.m_name} (will retry in {retry_delay}s because of 'unavailable' error; details: '{e.details()}')")
            time.sleep(retry_delay)
            self.logger.debug(f"<RPC> >> {self.s_name}.{self.m_name}... (retry)")
        else:
            # Timed out or any other reason: raise exception
            self.logger.debug(f"<RPC> >> {self.s_name}.{self.m_name} error: {str(e)}")
            raise RpcException(f"RPC error (on {trace}): {e}", rc=ResultCode.ERROR_RPC)

    def raise_result(self, result):
        if (
            (hasattr(result, "r") and isinstance(result.r, Result))
            and result.r.code > ResultCode.OK
            and result.r.code < ResultCode.ERROR_CUSTOM
            and self.exception
        ):
            # Error occurred
            raise self.custom_exception(f"RPC returned error: {result.r.msg}", rc=result.r.code)


class RetryStreamingMethod(RetryMethod):
    def __call__(self, request):
        # Call prelude
        trace, first_try = self.prelude(request)
        retry_delay = RPC_RETRY_DELAY

        # Loop to handle retries
        while True:
            try:
                # Call real stub method, with metadata
                for result in getattr(self.stub, self.m_name)(request, metadata=self.metadata.as_tuple()):
                    retry_delay = RPC_RETRY_DELAY
                    self.logger.debug(trace_rpc(False, result, context=self.metadata, method=f"{self.s_name}.{self.m_name}"))

                    # May raise an exception...
                    self.raise_result(result)
                    yield result
                break
            except RpcError as e:
                self.handle_exception(trace, first_try, e, retry_delay)
                retry_delay *= 2


class RetrySimpleMethod(RetryMethod):
    def __call__(self, request):
        # Call prelude
        trace, first_try = self.prelude(request)
        retry_delay = RPC_RETRY_DELAY

        # Loop to handle retries
        while True:
            try:
                # Call real stub method, with metadata
                result = getattr(self.stub, self.m_name)(request, metadata=self.metadata.as_tuple())
                self.logger.debug(trace_rpc(False, result, context=self.metadata, method=f"{self.s_name}.{self.m_name}"))

                # May raise an exception...
                self.raise_result(result)
                return result
            except RpcError as e:
                self.handle_exception(trace, first_try, e, retry_delay)
                retry_delay *= 2


# Utility class to handle stub retry
# i.e. permissive stub that allows server to be temporarily unavailable
class RetryStub:
    def __init__(self, real_stub, channel: str, timeout: float, metadata: RpcMetadata, logger: Logger, exception: bool, custom_exception: TypeVar):
        # Fake the stub methods
        for n in filter(lambda x: not x.startswith("__") and callable(getattr(real_stub, x)), dir(real_stub)):
            setattr(
                self,
                n,
                RetryStreamingMethod(n, real_stub, channel, timeout, metadata, logger, exception, custom_exception)
                if is_streaming(real_stub, n)
                else RetrySimpleMethod(n, real_stub, channel, timeout, metadata, logger, exception, custom_exception),
            )


class RpcClient:
    """
    Wrapper to GRPC api to setup an RPC client.

    Arguments:
        host:
            host string for RPC server to connect to.
        port:
            TCP port for RPC server to connect to.
        stubs_map:
            name:(stub,api_version) item maps, used to instantiate and attach (generated) stubs to current client instance
            e.g. if a {"foo": (FooServiceStub,1)} map is provided, methods of the Foo service can be accessed through a client.foo.xxx call
        timeout:
            timeout for unreachable server (in seconds; default: 60)
        name:
            can be either:
                - a string: client name, which will appear in server logs; RPC metadata will be generated from this name + current host
                - an RpcMetadata instance (when proxying calls from another client)
        logger:
            Logger instance to be used when logging calls
        exception:
            Raise RpcException if RPC includes a non-OK ResultCode status
        custom_exception:
            Use this provided custom class instead of RpcException when raising non-OK ResultCode exceptions
    """

    def __init__(
        self,
        host: str,
        port: int,
        stubs_map: dict,
        timeout: float = 60,
        name: Union[str, RpcMetadata] = "",
        logger: Logger = None,
        exception: bool = True,
        custom_exception: TypeVar = None,
    ):
        # Prepare metadata for RPC calls
        shared_metadata = name if isinstance(name, RpcMetadata) else RpcMetadata(name, self.get_user(), socket.gethostname(), get_current_ip())

        # Prepare logger
        self.logger = logger if logger is not None else getLogger("RpcClient")

        # Create channel
        self.target_host = f"{host}:{port}"
        self.logger.debug(f"Initializing RPC client for {self.target_host}")
        channel = insecure_channel(self.target_host)

        # Handle stubs hooking
        for name, typ_n_ver in stubs_map.items():
            typ, ver = typ_n_ver
            metadata = copy.copy(shared_metadata)
            if ver is not None:
                metadata.api_version = str(ver)
            self.logger.debug(f" >> adding {name} stub to client (api version: {ver})")
            setattr(self, name, RetryStub(typ(channel), self.target_host, timeout, metadata, self.logger, exception, custom_exception))

        self.logger.debug(f"RPC client ready for {self.target_host}")

    def get_user(self):
        if os.name == "posix":
            # Resolve user
            uid = os.getuid()
            try:
                # Try from pwd
                import pwd

                user = pwd.getpwuid(uid).pw_name
            except Exception:  # pragma: no cover
                # Not in pwd database, just keep UID
                user = f"{uid}"
            return user

        # Otherwise, just get login
        return os.getlogin()  # pragma: no cover



import logging
import os
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path

from grpc_helper.static_config import RpcStaticConfig


def add_rotating_handler(log_folder: Path, logger: logging.Logger):
    # Configure persisting folder/file for logs
    log_file = log_folder / logger.name / f"{logger.name}.log"
    log_file.parent.mkdir(parents=True, exist_ok=True)

    # Setup rotating handler with specific format
    handler = TimedRotatingFileHandler(
        log_file,
        when=RpcStaticConfig.LOGS_ROLLOVER_INTERVAL_UNIT.str_val,
        interval=RpcStaticConfig.LOGS_ROLLOVER_INTERVAL.int_val,
        backupCount=RpcStaticConfig.LOGS_BACKUP.int_val,
    )
    handler.setFormatter(
        logging.Formatter(
            f"%(asctime)s.%(msecs)03d [%(threadName)s{'/%(name)s' if isinstance(logger,logging.RootLogger) else ''}] %(levelname)s %(message)s - %(filename)s:%(funcName)s:%(lineno)d",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    logger.addHandler(handler)

    # First log!
    logger.info(f"---------- New {logger.name} logger instance for process {os.getpid()} ----------")


def clean_rotating_handler(logger):
    # Remove any rotating handler
    logger.info("Closing file log (shutting down)")
    for handler in filter(lambda h: isinstance(h, TimedRotatingFileHandler), logger.handlers):
        logger.removeHandler(handler)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from grpc_helper.api import common_pb2 as grpc__helper_dot_api_dot_common__pb2
from grpc_helper.api import config_pb2 as grpc__helper_dot_api_dot_config__pb2


class ConfigServiceStub(object):
    """Config service
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.get = channel.unary_unary(
                '/ConfigService/get',
                request_serializer=grpc__helper_dot_api_dot_common__pb2.Filter.SerializeToString,
                response_deserializer=grpc__helper_dot_api_dot_config__pb2.ConfigStatus.FromString,
                )
        self.reset = channel.unary_unary(
                '/ConfigService/reset',
                request_serializer=grpc__helper_dot_api_dot_common__pb2.Filter.SerializeToString,
                response_deserializer=grpc__helper_dot_api_dot_config__pb2.ConfigStatus.FromString,
                )
        self.set = channel.unary_unary(
                '/ConfigService/set',
                request_serializer=grpc__helper_dot_api_dot_config__pb2.ConfigUpdate.SerializeToString,
                response_deserializer=grpc__helper_dot_api_dot_config__pb2.ConfigStatus.FromString,
                )


class ConfigServiceServicer(object):
    """Config service
    """

    def get(self, request, context):
        """Get configuration items
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def reset(self, request, context):
        """Reset configuration items to their default value
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def set(self, request, context):
        """Update configuration items values
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ConfigServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'get': grpc.unary_unary_rpc_method_handler(
                    servicer.get,
                    request_deserializer=grpc__helper_dot_api_dot_common__pb2.Filter.FromString,
                    response_serializer=grpc__helper_dot_api_dot_config__pb2.ConfigStatus.SerializeToString,
            ),
            'reset': grpc.unary_unary_rpc_method_handler(
                    servicer.reset,
                    request_deserializer=grpc__helper_dot_api_dot_common__pb2.Filter.FromString,
                    response_serializer=grpc__helper_dot_api_dot_config__pb2.ConfigStatus.SerializeToString,
            ),
            'set': grpc.unary_unary_rpc_method_handler(
                    servicer.set,
                    request_deserializer=grpc__helper_dot_api_dot_config__pb2.ConfigUpdate.FromString,
                    response_serializer=grpc__helper_dot_api_dot_config__pb2.ConfigStatus.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ConfigService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ConfigService(object):
    """Config service
    """

    @staticmethod
    def get(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ConfigService/get',
            grpc__helper_dot_api_dot_common__pb2.Filter.SerializeToString,
            grpc__helper_dot_api_dot_config__pb2.ConfigStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def reset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ConfigService/reset',
            grpc__helper_dot_api_dot_common__pb2.Filter.SerializeToString,
            grpc__helper_dot_api_dot_config__pb2.ConfigStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def set(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ConfigService/set',
            grpc__helper_dot_api_dot_config__pb2.ConfigUpdate.SerializeToString,
            grpc__helper_dot_api_dot_config__pb2.ConfigStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from pathlib import Path


class Folders:
    """
    Folders management

    Attributes (may all be None):
        system: Path instance of the system folder (location where multi-users shared config is stored)
        user: Path instance of the user folder (location where multi-applications user config is stored)
        workspace: Path instance of the workspace folder (location where application config is stored)
    """

    def __init__(self, system: Path = None, user: Path = None, workspace: Path = None):
        self.__system = system
        self.__user = user
        self.__workspace = workspace

    @property
    def system(self) -> Path:
        """
        System folder
        """
        return self.__system

    @property
    def user(self) -> Path:
        """
        User folder
        """
        if self.__user is not None:
            self.__user.mkdir(parents=True, exist_ok=True)
        return self.__user

    @property
    def workspace(self) -> Path:
        """
        Workspace folder
        """
        if self.__workspace is not None:
            self.__workspace.mkdir(parents=True, exist_ok=True)
        return self.__workspace



import socket
from pathlib import Path

from grpc._channel import _StreamStreamMultiCallable, _UnaryStreamMultiCallable

from grpc_helper.meta import RpcMetadata

MAX_TRACE_BUFFER_LEN = 1024


def trace_buffer(buffer) -> str:
    # Build buffer name and content trace (stripped if too long)
    buffer_str = str(buffer).replace("\n", " ")
    buffer_str = buffer_str if len(buffer_str) < MAX_TRACE_BUFFER_LEN else buffer_str[0:MAX_TRACE_BUFFER_LEN] + "..."
    buffer_name = type(buffer).__name__
    return f"{buffer_name}{{{buffer_str}}}"


def trace_rpc(input_rpc: bool, buffer, context=None, method=None) -> tuple:
    buffer = trace_buffer(buffer)

    if isinstance(context, RpcMetadata):
        # Context is directly set as an RpcMetadata instance
        peer = str(context)
    else:
        # Build peer information from context metadata
        peer = str(RpcMetadata.from_context(context))

        # Get method name from context
        p = Path(context._rpc_event.call_details.method.decode("utf-8"))
        method = f"{p.parent.name}.{p.name}"

    # Build full trace string
    if input_rpc:
        return f"[RPC] {peer} >>> {method} ({buffer})"
    else:
        return f"[RPC] {peer} <<< {method}: {buffer}"


def is_streaming(stub: object, n: str):
    # Check if method is streaming output
    stub_method = getattr(stub, n)
    return isinstance(stub_method, _UnaryStreamMultiCallable) or isinstance(stub_method, _StreamStreamMultiCallable)


def get_current_ip(default: str = "127.0.0.1"):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(("8.8.8.8", 1))
        out = s.getsockname()[0]
    except Exception:  # pragma: no cover
        out = default
    finally:
        s.close()
    return out



import re
import sys
from argparse import Action, ArgumentParser, ArgumentTypeError, Namespace
from os.path import expanduser
from pathlib import Path

from argcomplete.completers import DirectoriesCompleter

from grpc_helper.folders import Folders

# Config item parameter syntax
CONFIG_ITEM_DEF_PATTERN = re.compile("([a-z][a-z0-9-]*)=(.*)")


def expanded_path(arg: str) -> Path:
    # Path with user expanded value
    return Path(expanduser(arg))


class ConfigDictAction(Action):
    """
    Custom action to store config items as a dict
    """

    def __call__(self, parser, namespace, values, option_string=None):
        # First, validate name=value syntax
        m = CONFIG_ITEM_DEF_PATTERN.match(values)
        if m is None:
            raise ArgumentTypeError(f"Invalid syntax for config item definition: {values}")
        name = m.group(1)
        value = m.group(2)

        # Get map (initialized by default arg)
        config_map = getattr(namespace, self.dest)

        # Store value in map
        config_map[name] = value


class RpcCliParser:
    """
    CLI arguments parser for RPC servers

    Constructor arguments:
        description:
            Description string to be displayed in help
        version:
            Version string to be displayed with -V option
    """

    def __init__(self, description: str, version: str = None):
        # Initialize parser
        self.__parser = ArgumentParser(description=description)

        # Add version command
        if version is not None:
            self.parser.add_argument("-V", "--version", action="version", version=version)

    @property
    def parser(self) -> ArgumentParser:
        return self.__parser

    def with_rpc_args(
        self, default_port: int = 54321, default_sys: str = "/etc/grpc_helper", default_usr: str = "~/.config/grpc_helper", default_wks: str = "./grpc_helper"
    ):
        # Folders
        self.parser.add_argument(
            "--system", type=expanded_path, action="store", default=expanded_path(default_sys), help=f"Override system folder (default: {default_sys})"
        ).completer = DirectoriesCompleter
        self.parser.add_argument(
            "--user", type=expanded_path, action="store", default=expanded_path(default_usr), help=f"Override user folder (default: {default_usr})"
        ).completer = DirectoriesCompleter
        self.parser.add_argument(
            "-w", "--workspace", type=expanded_path, action="store", default=expanded_path(default_wks), help=f"Workspace folder (default: {default_wks})"
        ).completer = DirectoriesCompleter

        # RPC server serving port
        self.parser.add_argument("-p", "--port", type=int, action="store", default=default_port, help=f"RPC server port (default: {default_port})")

        # Default config values
        self.parser.add_argument("-c", "--config", metavar="NAME=VALUE", action=ConfigDictAction, default={}, help="Override default configuration item value")

        return self

    def parse(self, input_args: list = None) -> Namespace:
        # By default, parse from CLI args
        input_args = input_args if input_args is not None else sys.argv[1:]

        # Parse!
        args = self.parser.parse_args(input_args)

        # Add folders definition
        setattr(args, "folders", Folders(args.system, args.user, args.workspace))  # NOQA: B010

        return args



from dataclasses import dataclass


@dataclass
class RpcMetadata:
    """
    Data class holding metadata related to an RPC call:
    * client:       Client name (client code)
    * user:         User name running the calling process
    * host:         Host name of the calling process
    * ip:           IP address of the calling process
    * api_version:  API version known to the caller
    """

    client: str = ""
    user: str = ""
    host: str = ""
    ip: str = ""
    api_version: str = ""

    def as_tuple(self) -> tuple:
        # Build tuple from set attributes
        return tuple(filter(lambda x: len(x[1]), self.__dict__.items()))

    @classmethod
    def from_context(cls, context):
        # Parse GRPC execution context
        out = cls()
        context_meta = {k: v for k, v in context.invocation_metadata()}
        for key in filter(lambda x: x in context_meta, out.__dict__.keys()):
            setattr(out, key, context_meta[key])
        return out

    def __str__(self):
        # String representation of metadata
        return (
            f"[{self.client if len(self.client) else 'unknown'}]"
            + f"{self.user if len(self.user) else 'unknown'}"
            + f"@{self.host if len(self.host) else 'unknown'}"
            + f"({self.ip if len(self.ip) else 'unknown'})"
            + f" api:{self.api_version if len(self.api_version) else 0}"
        )



import time
import traceback
from abc import ABC, abstractmethod
from logging import Logger, getLogger
from threading import Event as ThreadEvent
from threading import Thread
from typing import List

from grpc_helper.api import Event, EventFilter, EventInterrupt, ResultCode
from grpc_helper.client import RpcClient
from grpc_helper.errors import RpcException
from grpc_helper.static_config import RPC_RETRY_DELAY


class EventsListener(ABC):
    """
    Abstract event listener, with resume handling

    Arguments:
        client:
            Rpc client with an "events" stub
        names:
            Event names to listen to.
        client_id:
            Persisted events listening ID (if any)
        logger:
            Logger instance (optional)
    """

    def __init__(self, client: RpcClient, names: List[str], client_id: int = None, logger: Logger = None):

        # Init some stuff
        self.client = client
        self.names = names
        self.logger = logger if logger is not None else getLogger(EventsListener.__name__)
        self.client_id = client_id
        self.ready = ThreadEvent()

        # Prepare listening thread
        self.listening_t = Thread(target=self.__listen_to_events, daemon=True)
        self.listening_t.start()

    def __listen_to_events(self):
        # Listening loop
        retry_delay = RPC_RETRY_DELAY
        while True:
            try:
                for s in self.client.events.listen(EventFilter(client_id=self.client_id, names=self.names)):
                    # Success, restore retry delay
                    retry_delay = RPC_RETRY_DELAY

                    # Remember ID
                    if not self.ready.is_set():
                        self.client_id = s.client_id
                        self.logger.debug(f"Event listener #{self.client_id} is ready")
                        self.ready.set()

                    # Something to notify?
                    if len(s.event.name):
                        self.logger.debug(f">> Event listener #{self.client_id} on_event({s.event.name})")
                        self.on_event(s.event)
                        self.logger.debug(f"<< Event listener #{self.client_id} on_event({s.event.name})")

                # Listen loop normal exit: listening was interrupted
                break
            except Exception as e:
                error_trace = (
                    "event service restarting"
                    if isinstance(e, RpcException) and e.rc == ResultCode.ERROR_STREAM_SHUTDOWN
                    else f"{e}\n" + "".join(traceback.format_tb(e.__traceback__))
                )
                self.logger.error(f"Error occurred in event listener #{self.client_id} internal loop: {error_trace}")
                self.logger.warning(f"Retry in {retry_delay}s")
                time.sleep(retry_delay)
                retry_delay *= 2

    @abstractmethod
    def on_event(self, event: Event):  # pragma: no cover
        """
        To be implemented by sub-classes; will be called on any received event
        """
        pass

    def interrupt(self):
        """
        Interrupts listening loop
        """
        self.logger.debug(f">> Interrupting event listener #{self.client_id}")
        self.client.events.interrupt(EventInterrupt(client_id=self.client_id))
        self.listening_t.join()
        self.logger.debug(f"<< Interrupting event listener #{self.client_id}")



import logging
from pathlib import Path
from typing import Iterable, Tuple, Union

from grpc_helper.api import Filter, LoggerConfig, LoggerLevel, LoggerStatus, LoggerUpdate, ResultCode
from grpc_helper.api.logger_pb2_grpc import LoggerServiceServicer
from grpc_helper.errors import RpcException
from grpc_helper.manager import RpcManager

# Loggers config file name
LOGGERS_FILE = "loggers.json"


class LogsManager(LoggerServiceServicer, RpcManager):
    """
    Loggers manager, and implementing the LoggerService API
    """

    def __init__(self):
        RpcManager.__init__(self, LOGGERS_FILE, self.__validate_config_file)

        # Remember reset level for root logger
        self._root_reset_level = logging.getLogger().level

    def __validate_config_file(self, config_file: Path, json_model):
        if not isinstance(json_model, dict) or any(not isinstance(v, str) and not isinstance(v, bool) for v in json_model.values()):
            raise RpcException(f"Invalid logger json file (expecting a simple str:[str or bool] object): {config_file}", ResultCode.ERROR_MODEL_INVALID)

    def get_logger(self, name: str) -> logging.Logger:
        return logging.getLogger(name if len(name) else None)

    def _load(self):
        # Load loggers configuration, layer per layer
        loggers = {}

        # 1. System level
        loggers.update(self._load_config(self.folders.system))
        self.logger.debug(f"Loading loggers config (from system folder at {self.folders.system}): {loggers}")

        # 2. User level
        loggers.update(self._load_config(self.folders.user))
        self.logger.debug(f"Loading loggers config (from user folder at {self.folders.system}): {loggers}")

        # 3. Workspace level
        loggers.update(self._load_config(self.folders.workspace))
        self.logger.debug(f"Loading loggers config (from workspace folder at {self.folders.system}): {loggers}")

        # Update loggers
        for name, level in loggers.items():
            logger = self.get_logger(name)
            if isinstance(level, bool):
                # Global enablement
                logger.disabled = not level
            else:
                # Level configuration: read from string
                try:
                    logger.setLevel(LoggerLevel.Value(f"LVL_{level}"))
                except ValueError:
                    self.logger.warn(f"Ignoring unknown level {level} for logger {name}")

    def __inner_level_to_api(self, inner_level: int) -> LoggerLevel:
        for v in LoggerLevel.values():
            if inner_level == v:
                return v
        # Probably a custom level; this is unknown by public API
        return LoggerLevel.LVL_UNKNOWN

    def __map_loggers(self, request: Union[Filter, LoggerUpdate]) -> Iterable[Tuple[logging.Logger, LoggerConfig]]:
        # Don't support empty request
        if (isinstance(request, Filter) and len(request.names) == 0) or (isinstance(request, LoggerUpdate) and len(request.items) == 0):
            raise RpcException("Empty request", rc=ResultCode.ERROR_PARAM_MISSING)

        # Map to a tuple of (logger, config)
        if isinstance(request, Filter):
            # Provide reset config (enabled + default level)
            return map(
                lambda n: (self.get_logger(n), LoggerConfig(name=n, enabled=True, level=self._root_reset_level if n == "" else LoggerLevel.LVL_UNKNOWN)),
                request.names,
            )
        else:
            return map(lambda r: (self.get_logger(r.name), r), request.items)

    def __map_logger_config(self, logger: logging.Logger) -> LoggerConfig:
        return LoggerConfig(name=logger.name, enabled=not logger.disabled, level=self.__inner_level_to_api(logger.level))

    def get(self, request: Filter) -> LoggerStatus:
        with self.lock:
            # Get all loggers status
            out = LoggerStatus()
            for logger, _useless in self.__map_loggers(request):
                out.items.append(self.__map_logger_config(logger))
        return out

    def __generic_set(self, request: Union[Filter, LoggerUpdate], persist: bool):
        with self.lock:
            # Update all required loggers to provided configuration
            out = LoggerStatus()
            config = self._load_config(self.folders.workspace)
            config_updated = False
            for logger, l_config in self.__map_loggers(request):
                # Update logger from parameters
                logger.setLevel(l_config.level)
                logger.disabled = not l_config.enabled

                # Load config, ready to be updated
                persisted_name = "" if isinstance(logger, logging.RootLogger) else logger.name
                if persist:
                    # Persist new configuration
                    config[persisted_name] = False if logger.disabled else LoggerLevel.Name(logger.level).replace("LVL_", "")
                    config_updated = True
                elif persisted_name in config:
                    # Something was persisted, and has now to be removed
                    del config[persisted_name]
                    config_updated = True

                # Update output status
                out.items.append(self.__map_logger_config(logger))

            # Update configuration one time after loop (if something was modified)
            if config_updated:
                self._save_config(config)
        return out

    def set(self, request: LoggerUpdate) -> LoggerStatus:  # NOQA: A003
        return self.__generic_set(request, True)

    def reset(self, request: Filter) -> LoggerStatus:
        return self.__generic_set(request, False)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from grpc_helper.api import common_pb2 as grpc__helper_dot_api_dot_common__pb2
from grpc_helper.api import server_pb2 as grpc__helper_dot_api_dot_server__pb2


class RpcServerServiceStub(object):
    """Generic server behavior handling service
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.info = channel.unary_unary(
                '/RpcServerService/info',
                request_serializer=grpc__helper_dot_api_dot_common__pb2.Filter.SerializeToString,
                response_deserializer=grpc__helper_dot_api_dot_server__pb2.MultiServiceInfo.FromString,
                )
        self.shutdown = channel.unary_unary(
                '/RpcServerService/shutdown',
                request_serializer=grpc__helper_dot_api_dot_server__pb2.ShutdownRequest.SerializeToString,
                response_deserializer=grpc__helper_dot_api_dot_common__pb2.ResultStatus.FromString,
                )
        self.proxy_register = channel.unary_unary(
                '/RpcServerService/proxy_register',
                request_serializer=grpc__helper_dot_api_dot_server__pb2.ProxyRegisterRequest.SerializeToString,
                response_deserializer=grpc__helper_dot_api_dot_common__pb2.ResultStatus.FromString,
                )
        self.proxy_forget = channel.unary_unary(
                '/RpcServerService/proxy_forget',
                request_serializer=grpc__helper_dot_api_dot_common__pb2.Filter.SerializeToString,
                response_deserializer=grpc__helper_dot_api_dot_common__pb2.ResultStatus.FromString,
                )


class RpcServerServiceServicer(object):
    """Generic server behavior handling service
    """

    def info(self, request, context):
        """Get information
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def shutdown(self, request, context):
        """Graceful shutdown request
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def proxy_register(self, request, context):
        """Register proxy services
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def proxy_forget(self, request, context):
        """Forget previously registered proxy services
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RpcServerServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'info': grpc.unary_unary_rpc_method_handler(
                    servicer.info,
                    request_deserializer=grpc__helper_dot_api_dot_common__pb2.Filter.FromString,
                    response_serializer=grpc__helper_dot_api_dot_server__pb2.MultiServiceInfo.SerializeToString,
            ),
            'shutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.shutdown,
                    request_deserializer=grpc__helper_dot_api_dot_server__pb2.ShutdownRequest.FromString,
                    response_serializer=grpc__helper_dot_api_dot_common__pb2.ResultStatus.SerializeToString,
            ),
            'proxy_register': grpc.unary_unary_rpc_method_handler(
                    servicer.proxy_register,
                    request_deserializer=grpc__helper_dot_api_dot_server__pb2.ProxyRegisterRequest.FromString,
                    response_serializer=grpc__helper_dot_api_dot_common__pb2.ResultStatus.SerializeToString,
            ),
            'proxy_forget': grpc.unary_unary_rpc_method_handler(
                    servicer.proxy_forget,
                    request_deserializer=grpc__helper_dot_api_dot_common__pb2.Filter.FromString,
                    response_serializer=grpc__helper_dot_api_dot_common__pb2.ResultStatus.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'RpcServerService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RpcServerService(object):
    """Generic server behavior handling service
    """

    @staticmethod
    def info(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RpcServerService/info',
            grpc__helper_dot_api_dot_common__pb2.Filter.SerializeToString,
            grpc__helper_dot_api_dot_server__pb2.MultiServiceInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def shutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RpcServerService/shutdown',
            grpc__helper_dot_api_dot_server__pb2.ShutdownRequest.SerializeToString,
            grpc__helper_dot_api_dot_common__pb2.ResultStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def proxy_register(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RpcServerService/proxy_register',
            grpc__helper_dot_api_dot_server__pb2.ProxyRegisterRequest.SerializeToString,
            grpc__helper_dot_api_dot_common__pb2.ResultStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def proxy_forget(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RpcServerService/proxy_forget',
            grpc__helper_dot_api_dot_common__pb2.Filter.SerializeToString,
            grpc__helper_dot_api_dot_common__pb2.ResultStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import json
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from threading import RLock
from typing import Callable, List, Set, Tuple

from grpc_helper.api import EventApiVersion, Filter, ProxyRegisterRequest, ResultCode, ServerApiVersion
from grpc_helper.api.events_pb2_grpc import EventServiceStub
from grpc_helper.api.server_pb2_grpc import RpcServerServiceStub
from grpc_helper.client import RpcClient
from grpc_helper.errors import RpcException
from grpc_helper.folders import Folders
from grpc_helper.logs.logs_utils import add_rotating_handler
from grpc_helper.static_config import RpcStaticConfig
from grpc_helper.utils import get_current_ip


class RpcManager:
    """
    Shared interface for all RPC managers, helping to control the service lifecycle.
    """

    def __init__(self, config_name: str = None, config_validator: Callable = None):
        # Prepare some shared attributes
        self.lock = RLock()
        self.client = None
        self.folders = None
        self.logger = logging.getLogger(type(self).__name__)
        self.config_name = config_name
        self.config_validator = config_validator

    @property
    def _log_folder(self) -> Path:
        log_subfolder = Path(RpcStaticConfig.LOGS_FOLDER.str_val)
        log_folder = self.folders.workspace / log_subfolder if self.folders.workspace is not None else log_subfolder
        log_folder.parent.mkdir(parents=True, exist_ok=True)
        return log_folder

    def _init_folders_n_logger(self, folders: Folders, port: int):
        # Remember folders and port
        self.folders = folders if folders is not None else Folders()
        self.port = port

        # Prepare handler
        add_rotating_handler(self._log_folder, self.logger)

    def _preload(self, client: RpcClient, folders: Folders):
        # Remember client, and delegate loading to subclass
        self.client = client
        self._load()

    def _load(self):
        """
        To be defined by sub-classes, if some specific operations have to be performed once all basic initializations are done
        """
        pass

    def _shutdown(self):
        """
        To be defined by sub-classes, if some specific operations have to be performed before shutting down the process
        """

    def _load_config(self, config_folder: Path) -> dict:
        # Check config file presence
        if config_folder is not None and self.config_name is not None:
            config_file = config_folder / self.config_name
            if config_file.is_file():
                with config_file.open("r") as f:
                    # Load and verify model from json file
                    try:
                        json_model = json.load(f)
                    except json.JSONDecodeError as e:
                        raise RpcException(f"Invalid config json file (bad json: {e}): {config_file}", ResultCode.ERROR_MODEL_INVALID)

                    # Also validate model with provided validator
                    if self.config_validator is not None:
                        self.config_validator(config_file, json_model)

                    # Model looks to be valid: go on
                    return json_model

        # Default model
        return {}

    def _save_config(self, config: dict):
        # Save config file to workspace (if defined)
        folder = self.folders.workspace
        if folder is not None:
            folder.mkdir(parents=True, exist_ok=True)
            config_file = folder / self.config_name
            with config_file.open("w") as f:
                json.dump(config, f, indent=4)
        else:
            self.logger.warn("No workspace defined; skip config persistence")

    @property
    def _proxied_servers(self) -> Set[Tuple[str, int]]:
        # Tuples of remote RPC server host,port for each registered proxied service
        proxied_servers = set()
        for service_info in filter(lambda si: si.is_proxy and si.proxy_port > 0, self.client.srv.info(Filter()).items):
            proxied_servers.add((service_info.proxy_host if len(service_info.proxy_host) else RpcStaticConfig.MAIN_HOST.str_val, service_info.proxy_port))
        return proxied_servers

    def _proxied_clients(self, stubs_map: dict) -> List[RpcClient]:
        # Map proxied servers to clients
        out = []
        for host, port in self._proxied_servers:
            out.append(RpcClient(host, port, stubs_map, name=type(self).__name__, timeout=RpcStaticConfig.CLIENT_TIMEOUT.float_val, logger=self.logger))
        return out


class RpcProxiedManager(RpcManager, ABC):
    def __init__(self, config_name: str = None, config_validator: Callable = None):
        super().__init__(config_name=config_name, config_validator=config_validator)
        self.proxy_client = None

    def _proxy_stubs(self) -> dict:
        """
        Dict of stubs to be added to proxy_client
        """
        return {}

    @abstractmethod
    def _proxied_services(self) -> List[str]:  # pragma: no cover
        """
        List of service names to be proxied
        """
        pass

    @abstractmethod
    def _proxied_version(self) -> str:  # pragma: no cover
        """
        Version of proxied service
        """
        pass

    @property
    def _proxy_use_current_host(self) -> bool:
        """
        States if proxy registration should use current host IP or not (default)
        """
        return False

    def _load(self):
        # Prepare proxy client
        stubs_map = {"srv": (RpcServerServiceStub, ServerApiVersion.SERVER_API_CURRENT), "events": (EventServiceStub, EventApiVersion.EVENT_API_CURRENT)}
        stubs_map.update(self._proxy_stubs())
        client_candidate = RpcClient(
            RpcStaticConfig.MAIN_HOST.str_val, RpcStaticConfig.MAIN_PORT.int_val, stubs_map, timeout=RpcStaticConfig.CLIENT_TIMEOUT.float_val
        )

        # Register in proxy
        client_candidate.srv.proxy_register(
            ProxyRegisterRequest(
                names=self._proxied_services(),
                version=self._proxied_version(),
                host=self.__current_ip if self._proxy_use_current_host else RpcStaticConfig.MAIN_HOST.str_val,
                port=self.port,
            )
        )
        self.proxy_client = client_candidate

    @property
    def __current_ip(self):
        return get_current_ip(RpcStaticConfig.MAIN_HOST.str_val)

    def _shutdown(self):
        # Forget from proxy (if it was registered)
        if self.proxy_client is not None:
            self.proxy_client.srv.proxy_forget(Filter(names=self._proxied_services()))



import time
import traceback
from queue import Queue
from threading import Event as ThreadEvent
from threading import Thread

from grpc_helper.api import Empty, Event, EventFilter, EventInterrupt, EventQueueStatus, EventStatus, Result, ResultCode, ResultStatus
from grpc_helper.api.events_pb2_grpc import EventServiceServicer
from grpc_helper.errors import RpcException
from grpc_helper.folders import Folders
from grpc_helper.manager import RpcManager
from grpc_helper.static_config import RpcStaticConfig


class EventsManager(EventServiceServicer, RpcManager):
    """
    Events manager, and implementing the EventService API
    """

    def __init__(self):
        # Super call
        super().__init__("queues.json")

        # Some setup
        self.__queues = {}
        self.__interrupt_times = {}

        # Init keep alive thread
        self.__keep_alive_stop = ThreadEvent()
        self.__keep_alive_t = None

    def _init_folders_n_logger(self, folders: Folders, port: int):
        # Super call
        super()._init_folders_n_logger(folders, port)

        # Prepare persisted queues
        persisted_q = list(map(int, self._load_config(self.folders.workspace).keys()))
        self.logger.debug(f"Re-creating persisted event queues: {persisted_q}")
        for q_index in persisted_q:
            self.__queues[q_index] = Queue()

            # Also assume all queues are interrupted (i.e. will disappear if listen is not resumed within the retain timeout)
            self.__interrupt_times[q_index] = time.time()

    def _load(self):
        super()._load()

        # Start keep alive thread
        self.__keep_alive_t = Thread(target=self.__keep_alive, daemon=True)
        self.__keep_alive_t.start()

    def _persist_queues(self):
        with self.lock:
            # Persist queues
            self._save_config({i: [] for i in self.__queues.keys()})

    def _shutdown(self):
        with self.lock:
            # Stop keep alive thread
            if self.__keep_alive_t is not None:  # pragma: no branch
                self.__keep_alive_stop.set()
                self.__keep_alive_t.join()

            # Browse pending queues
            for q in self.__queues.values():
                # Push an interrupt status
                q.put(EventStatus(r=Result(msg="Service is shutdown", code=ResultCode.ERROR_STREAM_SHUTDOWN)), False)

    def __get_queue(self, index: int) -> Queue:
        with self.lock:
            q = self.__queues[index] if index in self.__queues else None
            if q is None:
                raise RpcException(f"Unknown event listening ID: {index}", ResultCode.ERROR_ITEM_UNKNOWN)
            return q

    def interrupt(self, request: EventInterrupt) -> ResultStatus:
        # Known index?
        q = self.__get_queue(request.client_id)

        # Refuse to interrupt again already interrupted queue
        with self.lock:
            if request.client_id in self.__interrupt_times and self.__interrupt_times[request.client_id] is not None:
                raise RpcException(f"Already interrupted ID: {request.client_id}", ResultCode.ERROR_STATE_UNEXPECTED)

        # Put "None" event (will wait until listen loop is over)
        q.put(None)
        return ResultStatus()

    def listen(self, request: EventFilter) -> EventStatus:
        # Validate filter
        for name in request.names:
            if len(name) == 0:
                raise RpcException("Empty name in filter request", ResultCode.ERROR_PARAM_MISSING)
            if " " in name:
                raise RpcException(f"Invalid name in filter request: {name}", ResultCode.ERROR_PARAM_INVALID)

        with self.lock:
            # Event queue specified?
            if request.client_id > 0:
                # Try to work with already existing queue
                index = request.client_id
                q = self.__get_queue(index)
                self.logger.info(f"Resuming interrupted event listening queue #{index}")

                # Listening resumed: forget interrupt time
                self.__interrupt_times[index] = None
            else:
                # Prepare a new index (from scratch, to reuse released indexes)
                index = 1
                while index in self.__queues:
                    index += 1
                self.logger.info(f"Starting new event listening queue #{index}")

                # Create new queue
                q = Queue()
                self.__queues[index] = q
                self._persist_queues()

        # Yield at least a first status with client_id
        yield EventStatus(client_id=index)

        # Loop to flush received events
        while True:
            # Blocking read
            event = q.get()

            # End of loop?
            if event is None or isinstance(event, EventStatus):
                break

            # Is this a meaningful or a keep alive event?
            if len(request.names) == 0 or event.name in request.names or event.name == "":
                # Yield received events (if any)
                yield EventStatus(client_id=index, event=event)

        # Forget queue
        with self.lock:
            # Remember interupt time
            interrupt_time = time.time()
            self.__interrupt_times[index] = interrupt_time
            self.logger.info(f"Event listening queue #{index} interrupted")

        # Last event to be yield (in case of shutdown)?
        if isinstance(event, EventStatus):
            yield event

    def send(self, request: Event) -> ResultStatus:
        # Just check event name is correct
        if len(request.name) == 0:
            raise RpcException("Empty event name", ResultCode.ERROR_PARAM_MISSING)
        if " " in request.name:
            raise RpcException(f"Invalid event name: {request.name}", ResultCode.ERROR_PARAM_INVALID)

        return self.__internal_send(request)

    def __internal_send(self, request: Event) -> ResultStatus:
        with self.lock:
            # Browse all queues
            to_delete = []
            for index in self.__queues.keys():
                # Reckon interrupted time
                interrupted_time = (
                    (time.time() - self.__interrupt_times[index]) if index in self.__interrupt_times and self.__interrupt_times[index] is not None else None
                )
                retain_timeout = RpcStaticConfig.EVENT_RETAIN_TIMEOUT.int_val

                # Interrupted queue + retain timeout expired?
                if interrupted_time is not None and interrupted_time >= retain_timeout:
                    # Yes, forget event queue
                    self.logger.info(f"Deleting interrupted event queue #{index} (retain timeout: {interrupted_time} >= {retain_timeout})")
                    to_delete.append(index)
                else:
                    # Still active queue: push event
                    self.__queues[index].put_nowait(request)

            # Forget timed out events (if any)
            if len(to_delete):
                for index in to_delete:
                    del self.__queues[index]
                    del self.__interrupt_times[index]
                self._persist_queues()

        return ResultStatus()

    def inspect(self, request: Empty) -> EventQueueStatus:
        # Just dump active queues
        with self.lock:
            return EventQueueStatus(client_ids=list(self.__queues.keys()))

    def __keep_alive(self):
        while not self.__keep_alive_stop.is_set():
            try:
                # Send keep alive event
                self.__internal_send(Event())

                # Sleep before looping
                init_time = time.time()
                while (not self.__keep_alive_stop.is_set()) and (time.time() - init_time) < RpcStaticConfig.EVENT_KEEPALIVE_TIMEOUT.int_val:
                    time.sleep(1)
            except Exception as e:  # pragma: no cover
                self.logger.error(f"Exception while sending keep alive event: {e}\n" + "".join(traceback.format_tb(e.__traceback__)))
